/*
 *  Copyright (c) 2020 NetApp
 *  All rights reserved
 */

package nvc

import (
	"context"
	"fmt"
	"net/http"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	log "github.com/sirupsen/logrus"
	"k8s.io/apimachinery/pkg/labels"

	//Quark
	corev1 "k8s.io/api/core/v1"
	"quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

var srv *http.Server

var (
	provisonedProtocol          = "ntapvol_protocol_provisioned"
	provisionedHelp             = "Total number of volumes based on protocol"
	provisionedLabels  []string = []string{"protocol", "storagePoolUUID"}
	desc               *prometheus.Desc
)

func (c *Controller) Describe(ch chan<- *prometheus.Desc) {
	// Describe new metric
	desc = prometheus.NewDesc(provisonedProtocol, provisionedHelp, provisionedLabels, nil)
	ch <- desc
}

func (c *Controller) Collect(ch chan<- prometheus.Metric) {
	// List existing storage pools
	sPools, err := c.storagePoolLister.List(labels.NewSelector())
	if err != nil {
		fmt.Errorf("error listing storage pool %v", err)
	}
	// For each storage pool, list volumes related to it.
	// and calculate number of nfs3 and cifs volume
	for _, sp := range sPools {
		nfs3, cifs := 0, 0
		ls, _ := labels.Parse("quark.netapp.io/storagePool=" + sp.Name)

		vols, err := c.volumesLister.List(ls)
		if err != nil {
			fmt.Errorf("error listing volumes %v", err)
			return
		}
		for _, vol := range vols {

			if isOnline(vol) {
				if vol.Spec.Protocol == "nfs3" {
					nfs3 = nfs3 + 1
				} else {
					cifs = cifs + 1
				}
			}

		}
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(nfs3),
			[]string{"nfs3", sp.Name}...,
		)
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(cifs),
			[]string{"cifs", sp.Name}...,
		)
	}
}

var (
	// Provision time for last volume by NVC
	volumeProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provision_time",
			Help:      "Total time spent provisioning last Netapp Volume.",
		},
	)

	averageProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_avg_provision_time",
			Help:      "Average time to provision Netapp Volumes.",
		},
	)

	provisionedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provisioned_count_total",
			Help:      "Total count of successfully provisioned Netapp Volumes.",
		},
	)

	failedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failed_volume_count_total",
			Help:      "Total count of Netapp Volumes failed to provision.",
		},
	)

	knownVolumesCount = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_known_volume_count_total",
			Help:      "Total count of Netapp Volumes known to Netapp Volume Controller.",
		},
	)

	failureCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failure_count_total",
			Help:      "Total difference in actual state and desired state of volumes for Netapp Volume Controller.",
		},
	)

	provisionedStoragePoolCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_provisioned_count_total",
			Help:      "Total count of successfully provisioned Netapp Storage Pools.",
		},
	)

	failedStoragePoolsCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_failed_storage_pool_count_total",
			Help:      "Total count of Netapp Storage Pools failed to provision.",
		},
	)

	knownStoragePoolsCount = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_known_storage_pool_count_total",
			Help:      "Total count of Netapp Storage Pools known to Netapp Volume Controller.",
		},
	)

	storagePoolProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_provision_time",
			Help:      "Total time spent provisioning last Netapp Storage Pool.",
		},
	)

	storagePoolAverageProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_avg_provision_time",
			Help:      "Average time to provision Netapp Storage Pools.",
		},
	)
)

func updateVolumeProvisionTimeSeconds(provisionTime int) {
	volumeProvisionTimeSeconds.Set(float64(provisionTime))
}

func updateAverageProvisionTimeSeconds(avgTime int) {
	averageProvisionTimeSeconds.Set(float64(avgTime))
}

func incFailedVolumesCount() {
	failedVolumesCount.Inc()
}

func incFailureCount() {
	failureCount.Inc()
}

func updateKnownVolumesCount(knownVolCount int) {
	knownVolumesCount.Set(float64(knownVolCount))
}

func incProvisionedVolumesCount() {
	provisionedVolumesCount.Inc()
}

func incFailedStoragePoolCount() {
	failedStoragePoolsCount.Inc()
}

func updateKnownStoragePoolCount(knownCount int) {
	knownStoragePoolsCount.Set(float64(knownCount))
}

func incProvisionedStoragePoolsCount() {
	provisionedStoragePoolCount.Inc()
}

func updateStoragePoolProvisionTimeSeconds(provisionTime int) {
	storagePoolProvisionTimeSeconds.Set(float64(provisionTime))
}

func updateStoragePoolAverageProvisionTimeSeconds(avgTime int) {
	storagePoolAverageProvisionTimeSeconds.Set(float64(avgTime))
}

func isOnline(volume *v1alpha1.NetAppVolume) bool {
	for _, cond := range volume.Status.Conditions {
		if cond.Type == v1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// StartMetricsServer will start metrics exporter for nvc
func StartMetricsServer() {

	prometheus.MustRegister(volumeProvisionTimeSeconds)
	prometheus.MustRegister(averageProvisionTimeSeconds)
	prometheus.MustRegister(provisionedVolumesCount)
	prometheus.MustRegister(failedVolumesCount)
	prometheus.MustRegister(knownVolumesCount)
	prometheus.MustRegister(failureCount)
	prometheus.MustRegister(provisionedStoragePoolCount)
	prometheus.MustRegister(failedStoragePoolsCount)
	prometheus.MustRegister(knownStoragePoolsCount)
	prometheus.MustRegister(storagePoolProvisionTimeSeconds)
	prometheus.MustRegister(storagePoolAverageProvisionTimeSeconds)

	server := http.NewServeMux()
	// Handle the endpoint serving the metrics
	server.Handle("/metrics", promhttp.Handler())
	// Create a server object
	srv = &http.Server{
		Addr:    ":" + quarkconfig.NvcMetricsPort,
		Handler: server,
	}

	go func() {
		if err := srv.ListenAndServe(); err != nil && err != http.ErrServerClosed {
			log.Fatalf("Error starting nvc exporter: %s\n", err)
		}
	}()
	log.Info("NVC exporter started")
}

// StopMetricsServer will stop metrics exporter when stop signal is received
func StopMetricsServer() {
	log.Info("NVC exporter shuting down")
	ctx, _ := context.WithTimeout(context.Background(), 5*time.Second)
	if err := srv.Shutdown(ctx); err != nil {
		log.Fatalf("NVC exporter shutdown Failed: %+v", err)
	}
}
