package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// NetAppClusterPeeringUpdate is used to signal the most recent state of a NetAppClusterPeering relation and the
// most recent changes.
type NetAppClusterPeeringUpdate struct {
	ClusterPeering    *quarkv1alpha1.NetAppClusterPeering
	UpdatedConditions []quarkv1alpha1.NetAppClusterPeeringCondition
}

// NetAppClusterPeeringDelete is used to signal the deletion of a NetAppClusterPeering relation
type NetAppClusterPeeringDelete struct {
	ClusterPeering *quarkv1alpha1.NetAppClusterPeering
}

func ConstructNetAppClusterPeering(
	ownerID string,
	clusterPeeringUUID string,
	storagePoolUUID, remoteStoragePoolUUID string,
	remoteNamespace, remoteLocation, remoteClusterName string,
	enableCCPd, enableServiceMesh bool,
) (clusterPeering *quarkv1alpha1.NetAppClusterPeering, err error) {

	// validate uuids
	_, err = uuid.FromString(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid uuid value for clusterPeeringUUID: %v", err)
	}
	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid uuid value for storagePoolUUID: %v", err)
	}
	_, err = uuid.FromString(remoteStoragePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid uuid value for remoteStoragePoolUUID: %v", err)
	}

	// Validate label values
	if len(strings.TrimSpace(remoteLocation)) == 0 ||
		len(strings.TrimSpace(ownerID)) == 0 ||
		len(strings.TrimSpace(remoteNamespace)) == 0 {
		return nil, fmt.Errorf("invalid input. One or more of the following values are not populated: %s: \"%s\", %s: \"%s\", %s: \"%s\"",
			"remoteLocation", remoteLocation, "ownerID", ownerID, "remoteNamespace", remoteNamespace,
		)
	}

	remoteService := quarkconfig.GetClusterPeeringServiceName(remoteStoragePoolUUID)
	remoteEndpoint := remoteService + "." + remoteNamespace + ".svc.cluster.local"

	finalizers := []string{}

	if enableCCPd {
		// This finalizer is managed by CCPd and includes cleaning up SMF tables and
		// any further invalidations needed when removing a peering for a storage pool.
		finalizers = append(finalizers, quarkconfig.ClusterPeeringFinalizer)
	}
	if enableServiceMesh {
		// If a service mesh is involved, we need to add a finalizer to manage cleanup of
		// service mesh resources.
		finalizers = append(finalizers, quarkconfig.ServiceMeshFinalizer)
	}

	// Create the object once the inputs are validated
	clusterPeering = &quarkv1alpha1.NetAppClusterPeering{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: clusterPeeringUUID,
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:           ownerID,
				quarkconfig.StoragePoolLabel:       storagePoolUUID,
				quarkconfig.RemoteStoragePoolLabel: remoteStoragePoolUUID,
				quarkconfig.RemoteLocationLabel:    remoteLocation,
			},
			Finalizers: finalizers,
		},
		Spec: quarkv1alpha1.NetAppClusterPeeringSpec{
			RemoteEndpoint: remoteEndpoint,
		},
	}
	if enableServiceMesh {
		clusterPeering.Labels[quarkconfig.RemoteClusterNameLabel] = remoteClusterName
	}

	return
}

// Makes API call to create the cluster-peering resource.
// Verify the referenced object/s (storage pool) exist on the k8s cluster.
func (client *NVCClient) NetAppClusterPeeringCreate(clusterPeering *quarkv1alpha1.NetAppClusterPeering) (*quarkv1alpha1.NetAppClusterPeering, error) {
	// Fetch storage pool to use for owner-reference, and also validate storagepool with this uuid exists
	if _, err := client.NetAppStoragePoolGet(clusterPeering.Labels[quarkconfig.StoragePoolLabel]); err != nil {
		return nil, fmt.Errorf("cannot create NetAppClusterPeering. Failed to verify if NetAppStoragePool with uuid %v exists. %v", clusterPeering.Labels[quarkconfig.StoragePoolLabel], err)
	}

	existingCPs, err := client.NetAppClusterPeeringGetAll()
	if err != nil {
		return nil, fmt.Errorf("cannot create NetAppClusterPeering. Failed to read existing NetAppClusterPeering: %v", err)
	}

	localSP := clusterPeering.Labels[quarkconfig.StoragePoolLabel]
	remoteSP := clusterPeering.Labels[quarkconfig.RemoteStoragePoolLabel]
	for _, existingCP := range existingCPs {
		if existingCP.Labels[quarkconfig.StoragePoolLabel] == localSP &&
			existingCP.Labels[quarkconfig.RemoteStoragePoolLabel] == remoteSP {
			return nil, fmt.Errorf("cannot create NetAppClusterPeering. NetAppClusterPeering %v already exists between NetAppStoragePool %v and remote NetAppStoragePool %v",
				existingCP.Name, localSP, remoteSP)
		}
	}

	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Create(
		context.TODO(), clusterPeering, metav1.CreateOptions{})
}

func (client *NVCClient) NetAppClusterPeeringDelete(clusterPeeringUUID string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Delete(
		context.TODO(), clusterPeeringUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppClusterPeeringGet(clusterPeeringUUID string) (*quarkv1alpha1.NetAppClusterPeering, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Get(
		context.TODO(), clusterPeeringUUID, metav1.GetOptions{})
}

func (client *NVCClient) NetAppClusterPeeringGetAll() ([]*quarkv1alpha1.NetAppClusterPeering, error) {
	clusterPeeringsList, err := client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	clusterPeerings := make([]*quarkv1alpha1.NetAppClusterPeering, len(clusterPeeringsList.Items))
	for i := range clusterPeeringsList.Items {
		clusterPeerings[i] = &clusterPeeringsList.Items[i]
	}
	return clusterPeerings, nil
}

func (client *NVCClient) NetAppClusterPeeringFilterByStoragePool(storagePoolName string) (*quarkv1alpha1.NetAppClusterPeeringList, error) {
	return client.NetAppClusterPeeringFilterByLabel(fmt.Sprintf("%s=%s", quarkconfig.StoragePoolLabel, storagePoolName))
}

func (client *NVCClient) NetAppClusterPeeringFilterByLabel(label string) (*quarkv1alpha1.NetAppClusterPeeringList, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{LabelSelector: label})
}

func NetAppClusterPeeringReady(clusterPeering *quarkv1alpha1.NetAppClusterPeering) bool {
	return (clusterPeering.Status.State == quarkv1alpha1.ClusterPeeringStatePeered)
}

func NetAppServiceMeshEnabled(clusterPeering *quarkv1alpha1.NetAppClusterPeering) bool {
	for _, cond := range clusterPeering.Status.Conditions {
		if cond.Type == quarkv1alpha1.ServiceMeshEnabled && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}
