package v2

import (
	"context"
	"fmt"
	"strings"

	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"

	_ "k8s.io/client-go/plugin/pkg/client/auth"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"

	"quark.netapp.io/volume-controller/pkg/nvc"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

var (
	// supporting smaller storagePool for upgrade and internal configs
	minStoragePoolSize      resource.Quantity = resource.MustParse("1Gi")
	maxStoragePoolSize      resource.Quantity = resource.MustParse("200Ti")
	maxStoragePoolSizeAzure resource.Quantity = resource.MustParse("100Ti")
)

const (
	versionSPUpgradeErr = "Failed to upgrade NetAppStoragePool"
	upgradeSPAbortErr   = "Failed to abort NetAppStoragePoolUpgrade"
	CMEKConfigMapErr    = "Failed to create cmek config map: %s not provided"
)

func isValidCMEKLogLevel(level string) bool {
	logLevel := strings.ToLower(level)
	var CMEKLogLevels = []string{"debug", "alert", "info", "critical", "error"}
	for _, value := range CMEKLogLevels {
		if value == logLevel {
			return true
		}
	}
	return false
}

type MigrateZoneAbortZoneFailureError struct {
}

func (e MigrateZoneAbortZoneFailureError) Error() string {
	return "Unable to abort migration caused by zone failure."
}

type MigrateZoneAbortNoMigrationFailureError struct {
}

func (e MigrateZoneAbortNoMigrationFailureError) Error() string {
	return "No zone migration to abort."
}

type GlobalAccessUpdateAbortNoUpdateError struct{}

func (e GlobalAccessUpdateAbortNoUpdateError) Error() string {
	return "global access update not in progress"
}

type GlobalAccessAbortInProgressError struct{}

func (e GlobalAccessAbortInProgressError) Error() string {
	return "global access abort in progress"
}

// NetAppStoragePoolOnline returns "true" when the Quark pod is ready to serve data.
func (client *NVCClient) NetAppStoragePoolOnline(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	for _, cond := range storagePool.Status.Conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// NetAppStoragePoolSMBConfigured returns "true" if SMB server id configured in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBConfigured(storagePool)
}

// NetAppStoragePoolSMBReady Returns "true" if SMB is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBConfigured(storagePool) && util.IsSMBServerConfigurationEqual(storagePool)
}

// NetAppStoragePoolSMBInitialized Returns "true" if SMB is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolSMBInitialized(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsStoragePoolSMBInitialized(storagePool)
}

// NetAppStoragePoolLdapReady Returns "true" if LDAP is configured and ready for use in storagepool.
func (client *NVCClient) NetAppStoragePoolLdapReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if !client.EnableCCPd || !client.EnableLdap {
		return false
	}

	return (client.NetAppStoragePoolSMBReady(storagePool) && util.IsStoragePoolLdapReady(storagePool))
}

// Get the storage pool status
func (client *NVCClient) NetAppStoragePoolStatus(storagePoolUUID string) (bool, string) {
	storagePoolLifeCycleState := false
	storagePoolLifeCycleStateDetails := "Storagepool is in error state"
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		storagePoolLifeCycleStateDetails = "Storagepool doesn't exists"
	} else if err != nil {
		storagePoolLifeCycleStateDetails = "Storagepool get failed"
	} else if client.NetAppStoragePoolOnline(storagePool) {
		storagePoolLifeCycleState = true
		storagePoolLifeCycleStateDetails = "Storagepool is available"
	}
	return storagePoolLifeCycleState, storagePoolLifeCycleStateDetails
}

// Initiate NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgrade(storagePoolUUID, version string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetappStoragePool %s was not found for upgrade", storagePoolUUID)
	}
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("%s: %s", versionSPUpgradeErr, err)
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("%s: no quarks found", versionSPUpgradeErr)
	}
	quark := quarks[0]
	if quark.Status.Version != version {
		return nil, fmt.Errorf("%s: quark is not updated to version %s", versionSPUpgradeErr, version)
	}

	// initiate upgrade when Quark is Ready and updated with required version for storagePool upgrade
	storagePool.Spec.QuarkVersion = version
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)

	return newStoragePool, err
}

// Abort NetAppStoragePool version upgrade
func (client *NVCClient) NetAppStoragePoolUpgradeAbort(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil && errors.IsNotFound(err) {
		return nil, fmt.Errorf("NetAppStoragePool %s was not found for aborting upgrade", storagePoolUUID)
	}
	if storagePool.Spec.QuarkVersion == "" {
		return nil, fmt.Errorf("%s: NetAppStoragePool Spec version is missing", upgradeSPAbortErr)
	}
	// check if abort is possible
	if storagePool.Status.QuarkVersion != "" && storagePool.Spec.QuarkVersion != storagePool.Status.QuarkVersion {
		err := client.checkUpgradeAbortStoragePool(storagePool.Status.Conditions)
		if err != nil {
			return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
		}
	}
	// revert to old version, in order to signal abort
	storagePool.Spec.QuarkVersion = storagePool.Status.QuarkVersion
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		_, staleError := err.(StaleVersionError)
		if staleError {
			return nil, err
		}
		return nil, fmt.Errorf("%s: %s", upgradeSPAbortErr, err.Error())
	}
	return storagePool, nil
}

// returns error if abort cannot proceed
func (client *NVCClient) checkUpgradeAbortStoragePool(conditions []quarkv1alpha1.NetAppStoragePoolCondition) error {
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded {
			// do not proceed with abort if halt succeeded
			return fmt.Errorf("deployment is being updated for storagePool upgrade")
		}
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolGet(storagePoolUUID string) (*quarkv1alpha1.NetAppStoragePool, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePoolUUID, metav1.GetOptions{})
}

func (client *NVCClient) NetAppStoragePoolGetVersion(storagePoolUUID string) (string, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return "", fmt.Errorf("failed to get the NetAppStoragePool version: %s", err.Error())
	}
	return storagePool.Status.QuarkVersion, nil
}

// returns conditions from NetAppStoragePool CR status
func (client *NVCClient) NetAppStoragePoolGetConditions(storagePoolUUID string) ([]quarkv1alpha1.NetAppStoragePoolCondition, error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	return storagePool.Status.Conditions, nil
}

func (client *NVCClient) NetAppStoragePoolGetAll() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	storagePoolsList, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range storagePoolsList.Items {
		storagePools = append(storagePools, &storagePoolsList.Items[i])
	}
	return storagePools, err
}

func (client *NVCClient) NetAppStoragePoolGetAvailableSpace(storagePoolUUID string) (availableSpace int64, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return availableSpace, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return availableSpace, fmt.Errorf("NetAppStoragePool %s is not online yet", storagePoolUUID)
	}
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return availableSpace, err
	}
	var usedStoragePoolSize int64 = 0
	for _, volume := range volumes.Items {
		usedStoragePoolSize = usedStoragePoolSize + volume.Status.Size.Value()
	}
	availableSpace = storagePool.Status.Size.Value() - usedStoragePoolSize
	return availableSpace, nil
}

func (client *NVCClient) NetAppStoragePoolGetSMBStatus(storagePoolUUID string) (configured bool, err error) {
	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil || storagePool == nil {
		return false, fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}
	if !client.NetAppStoragePoolOnline(storagePool) {
		return false, fmt.Errorf("NetAppStoragePool %s is not online yet", storagePoolUUID)
	}
	return client.NetAppStoragePoolSMBConfigured(storagePool), nil
}

func (client *NVCClient) NetAppStoragePoolGetAllSMB() ([]*quarkv1alpha1.NetAppStoragePool, error) {
	storagePools := make([]*quarkv1alpha1.NetAppStoragePool, 0)
	storagePoolsList, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	for i := range storagePoolsList.Items {
		if storagePoolsList.Items[i].Spec.SMBServerConfiguration.Netbios != "" {
			storagePools = append(storagePools, &storagePoolsList.Items[i])
		}
	}
	return storagePools, err
}

func (client *NVCClient) NetAppVolumeListGetByStoragePool(storagePoolUUID string) (*quarkv1alpha1.NetAppVolumeList, error) {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID
	labelSelector := fmt.Sprintf("quark.netapp.io/storagePool=%s", storagePoolUUID)
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}
	volumesList, err := client.quarkClient.QuarkV1alpha1().NetAppVolumes(client.nvcNamespace).List(
		context.TODO(), listOptions)
	if err != nil {
		return nil, err
	}
	return volumesList, err
}

func (client *NVCClient) NetAppStoragePoolDelete(storagePoolUUID string, user, password string) error {
	volumes, err := client.NetAppVolumeListGetByStoragePool(storagePoolUUID)
	if err != nil {
		return err
	}
	if len(volumes.Items) > 0 {
		return fmt.Errorf("NetAppStoragePool delete request denied, still in use")
	}

	storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
	}

	// Check if this storage pool is in any peering relationships and prevent deletion if it is.
	// (EnableCCPd should be true, else peering CRD won't be available on the cluster)
	if client.EnableCCPd {
		if clusterPeeringsList, err := client.NetAppClusterPeeringFilterByStoragePool(storagePool.Name); err != nil {
			return fmt.Errorf("failed to list NetAppClusterPeerings to check if storage pool %v is in a peering relationship: %v", storagePool.Name, err.Error())
		} else {
			if len(clusterPeeringsList.Items) > 0 {
				return fmt.Errorf(`delete request denied. NetAppStoragePool %v is in a peering relationship. Break/remove the peering relationship by deleting
					NetAppClusterPeering %v, then retry NetAppStoragePool delete.`, storagePool.Name, clusterPeeringsList.Items[0].Name)
			}
		}
	}

	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		storagePool, err := client.NetAppStoragePoolGet(storagePoolUUID)
		if err != nil {
			return fmt.Errorf("failed to get the NetAppStoragePool: %s", err.Error())
		}
		_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, util.GetVolDeleteADSecretName(storagePool.Name), user, password)
		if err != nil {
			return err
		}
	}
	return client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Delete(context.TODO(), storagePoolUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) netAppStoragePoolUpdate(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Update(context.TODO(), storagePool, metav1.UpdateOptions{})
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}
	return storagePool, nil
}

// NetAppStoragePoolResizeAbort aborts the current StoragePool-Resize operation
func (client *NVCClient) NetAppStoragePoolResizeAbort(storagePoolUUID string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	// Reset the new size in spec. The storagePool error conditions can then be cleared by Podrick.
	storagePool.Spec.Size = storagePool.Status.Size
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// Note that if the migration has completed by the time this operation is
// issued, the operation will do nothing
func (client *NVCClient) NetAppStoragePoolMigrateZoneAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(
		client.nvcNamespace).Get(
		context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	if storagePool.Spec.Zone == storagePool.Status.Zone {
		return nil, MigrateZoneAbortNoMigrationFailureError{}
	}

	if storagePool.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseFailure {
		// return an error here, since we can't abort zone migrations due to failure
		return nil, MigrateZoneAbortZoneFailureError{}
	}
	storagePool.Spec.Zone = storagePool.Status.Zone
	storagePool.Spec.LastZoneMigrationCause = storagePool.Status.LastZoneMigrationCause
	storagePool.Spec.ZoneMigrationCount = storagePool.Status.ZoneMigrationCount
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)
	return storagePool, err
}

// NetAppStoragePoolResized returns "true" when the data PVCs are correctly resized.
func (client *NVCClient) NetAppStoragePoolResized(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return sp.Status.Size.Cmp(sp.Spec.Size) == 0
}

func (client *NVCClient) NetAppStoragePoolMigratedZone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	// Only acknowledge a zone migration if it wasn't initiated as part of a zone
	// failure.
	return sp.Status.Zone == sp.Spec.Zone && sp.Spec.LastZoneMigrationCause == quarkconfig.ZoneMigrationCauseUser
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdated(
	sp *quarkv1alpha1.NetAppStoragePool,
) bool {
	// If global access enablement is being aborted, the spec and status fields
	// will be equal, but GlobalAccessAborting will be true. We only want to
	// declare this to be finalized once GlobalAccessAborting is set to false.
	return sp.Status.GlobalAccess == sp.Spec.GlobalAccess &&
		!sp.Spec.GlobalAccessAborting
}

// NetAppStoragePoolADDNSUpdated returns "true" when AD DNSIPs are applied to the storagePool.
func (client *NVCClient) NetAppStoragePoolADDNSUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsEqualStrings(sp.Spec.ADDNSIPs, sp.Status.ADDNSIPs)
}

// NetAppStoragePoolADPasswordResetDone returns "true" when machine-account password reset is successful.
func (client *NVCClient) NetAppStoragePoolADPasswordResetDone(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBPasswordResetCount == sp.Spec.SMBPasswordResetCount)
}

// NetAppStoragePoolSMBADConfigurationUpdated returns "true" when AD configurations updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return (sp.Status.SMBADConfiguration == sp.Spec.SMBADConfiguration)
}

// NetAppStoragePoolServerConfigurationUpdated returns "true" when SMB server configurations are updated for the storagePool.
func (client *NVCClient) NetAppStoragePoolSMBServerConfigurationUpdated(sp *quarkv1alpha1.NetAppStoragePool) bool {
	return util.IsSMBServerConfigurationEqual(sp)
}

// NetAppStoragePoolSMBADUserListUpdated returns "true" when all users are updated for a given storagePool.
func (client *NVCClient) NetAppStoragePoolSMBADUserListUpdated(sp *quarkv1alpha1.NetAppStoragePool) (ret bool) {
	if !client.EnableCCPd || !client.EnableBackupSecurityUsers {
		return false
	}
	val := (util.IsEqualStrings(sp.Spec.SMBBackupUsers, sp.Status.SMBBackupUsers) &&
		util.IsEqualStrings(sp.Spec.SMBSecurityPrivilegeUsers, sp.Status.SMBSecurityPrivilegeUsers))

	return val
}

// NetAppStoragePoolUpgraded returns "true" when StoragePool version upgrade is successful.
func (client *NVCClient) NetAppStoragePoolUpgraded(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	conditions := storagePool.Status.Conditions
	for _, cond := range conditions {
		if cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltSucceeded || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeDMAPHaltRequested || cond.Type == quarkv1alpha1.NetAppStoragePoolUpgradeFailed {
			// return false if another upgrade is in progress
			return false
		}
	}
	return (storagePool.Status.QuarkVersion == storagePool.Spec.QuarkVersion)

}

// NetAppStoragePoolDelete is used to signal the deletion of a NetAppStoragePool.
type NetAppStoragePoolDelete struct {
	StoragePool *quarkv1alpha1.NetAppStoragePool
}

// NetAppStoragePoolUpdate is used to signal the most recent state of a NetAppStoragePool and the
// most recent changes.
type NetAppStoragePoolUpdate struct {
	StoragePool            *quarkv1alpha1.NetAppStoragePool
	Status                 *NetAppStoragePoolUpdateStatus
	UpdatedConditions      []quarkv1alpha1.NetAppStoragePoolCondition
	LastZoneMigrationCause string
	Error                  error
}

// NetAppStoragePoolUpdateStatus captures the status of any pending updates.
type NetAppStoragePoolUpdateStatus struct {
	SizeUpdated                   bool
	ZoneUpdated                   bool
	GlobalAccessUpdated           bool
	ADDNSUpdated                  bool
	SMBADConfigurationUpdated     bool
	SMBADConfigUserListUpdated    bool
	SMBServerConfigurationUpdated bool
	ADPasswordReset               bool
	QuarkVersionUpdated           bool
	LdapConfigured                bool
}

// NetAppStoragePoolUpdated returns "true" when the actual state has converged to the
// desired state. If not, it will return an object describing what actions are pending.
func (client *NVCClient) NetAppStoragePoolUpdated(storagePool *quarkv1alpha1.NetAppStoragePool) (bool, *NetAppStoragePoolUpdateStatus, error) {
	statusConverged := true

	status := &NetAppStoragePoolUpdateStatus{
		SizeUpdated:                   true,
		ZoneUpdated:                   true,
		GlobalAccessUpdated:           true,
		ADDNSUpdated:                  true,
		SMBADConfigurationUpdated:     true,
		SMBADConfigUserListUpdated:    true,
		SMBServerConfigurationUpdated: true,
		ADPasswordReset:               true,
		QuarkVersionUpdated:           true,
		LdapConfigured:                true,
	}
	// Check size
	if !client.NetAppStoragePoolResized(storagePool) {
		status.SizeUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolMigratedZone(storagePool) {
		status.ZoneUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolGlobalAccessUpdated(storagePool) {
		status.GlobalAccessUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADDNSUpdated(storagePool) {
		status.ADDNSUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBADConfigurationUpdated(storagePool) {
		status.SMBADConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBADUserListUpdated(storagePool) {
		status.SMBADConfigUserListUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolLdapReady(storagePool) {
		status.LdapConfigured = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolSMBServerConfigurationUpdated(storagePool) {
		status.SMBServerConfigurationUpdated = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolADPasswordResetDone(storagePool) {
		status.ADPasswordReset = false
		statusConverged = false
	}
	if !client.NetAppStoragePoolUpgraded(storagePool) {
		status.QuarkVersionUpdated = false
		statusConverged = false
	}
	// Return the status of any potentially pending updates.
	if statusConverged {
		return true, status, nil
	} else {
		return false, status, nil
	}
}

func (client *NVCClient) handleStoragePoolDeletes(sp *quarkv1alpha1.NetAppStoragePool) {
	client.storagePoolDeleteChannel <- &NetAppStoragePoolDelete{sp}
}

func (client *NVCClient) handleStoragePoolUpdates(oldSp *quarkv1alpha1.NetAppStoragePool, newSp *quarkv1alpha1.NetAppStoragePool) {
	sendEvent := false

	// We send events when any of the NetAppStoragePool has converged to the desired state.
	_, newStatus, _ := client.NetAppStoragePoolUpdated(newSp)
	_, oldStatus, _ := client.NetAppStoragePoolUpdated(oldSp)
	if newStatus != oldStatus {
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"newStatus":   newStatus,
		}).Debug("NetApp Storage pool state has been updated.")
		sendEvent = true
	}
	// We send events when the NetStoragePool's conditions have changed.
	if len(newSp.Status.Conditions) != len(oldSp.Status.Conditions) {
		sendEvent = true
		log.WithFields(log.Fields{
			"storagePool": newSp.Name,
			"conditions":  newSp.Status.Conditions,
		}).Debug("NetApp Storage pool has updated conditions.")
	} else {
		for i, cond := range newSp.Status.Conditions {
			if oldSp.Status.Conditions[i].Type != cond.Type || oldSp.Status.Conditions[i].Status != cond.Status {
				sendEvent = true
				log.WithFields(log.Fields{
					"storagePool": newSp.Name,
					"conditions":  newSp.Status.Conditions,
				}).Debug("NetApp Storagepool has updated conditions.")
				break
			}
		}
	}
	if newSp.Spec.LastZoneMigrationCause != oldSp.Spec.LastZoneMigrationCause {
		sendEvent = true
	}

	if sendEvent {
		client.storagePoolUpdateChannel <- &NetAppStoragePoolUpdate{
			StoragePool:            newSp,
			Status:                 newStatus,
			UpdatedConditions:      newSp.Status.Conditions,
			LastZoneMigrationCause: newSp.Spec.LastZoneMigrationCause,
		}
	}
}

func (client *NVCClient) validateStoragePoolSize(size resource.Quantity) (err error) {
	var (
		maxStoragePoolSizeAllowed resource.Quantity
	)
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	if quark.Spec.Cloud == quarkconfig.CloudAzure {
		maxStoragePoolSizeAllowed = maxStoragePoolSizeAzure
	} else {
		maxStoragePoolSizeAllowed = maxStoragePoolSize
	}
	minStoragePoolSizeDifference := size.Value() - minStoragePoolSize.Value()
	maxStoragePoolSizeDifference := size.Value() - maxStoragePoolSizeAllowed.Value()
	if minStoragePoolSizeDifference < 0 || maxStoragePoolSizeDifference > 0 || minStoragePoolSizeDifference%incrementalSize.Value() != 0 {
		return fmt.Errorf("size=%s, storage pool size should be a minimum of %s, maximum of %s and a multiple of %s", size.String(), minStoragePoolSize.String(), maxStoragePoolSizeAllowed.String(), incrementalSize.String())
	}
	return nil
}

func (client *NVCClient) NetAppStoragePoolSizeUpdate(sp *quarkv1alpha1.NetAppStoragePool, newSizeString string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	size, err := resource.ParseQuantity(newSizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}

	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}
	if sp.Spec.Size.Value() > size.Value() {
		return nil, fmt.Errorf("storage pool size shrink operation is not allowed. Current storagePool size is %s, newsize=%s", sp.Spec.Size.String(), size.String())
	}
	sp.Spec.Size = size

	sp, err = client.netAppStoragePoolUpdate(sp)

	return sp, err
}

func (client *NVCClient) NetAppStoragePoolMigrateZone(
	sp *quarkv1alpha1.NetAppStoragePool, zone string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.Zone == zone {
		// We already did the migration; don't update.
		return sp, nil
	}
	// Validate zone choice
	zoneOk := false
	if len(sp.Status.AvailableZones) == 0 {
		return nil, fmt.Errorf("unable to migrate; no AvailableZones to choose from")
	}

	for i := 0; i < len(sp.Status.AvailableZones) && !zoneOk; i++ {
		zoneOk = zone == sp.Status.AvailableZones[i]
	}
	if !zoneOk {
		return nil, fmt.Errorf("zone %s is not in AvailableZones, choose one of %v", zone, sp.Status.AvailableZones)
	}

	sp.Spec.Zone = zone
	sp.Spec.LastZoneMigrationCause = quarkconfig.ZoneMigrationCauseUser
	sp.Spec.ZoneMigrationCount = sp.Spec.ZoneMigrationCount + 1
	// This is subject to clock skew, but there's not a good way around this.
	now := metav1.Now()
	sp.Spec.LastZoneMigrationTime = &now
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdate(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if sp.Spec.GlobalAccessAborting {
		return nil, GlobalAccessAbortInProgressError{}
	}
	sp.Spec.GlobalAccess = enabled
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessSet(
	sp *quarkv1alpha1.NetAppStoragePool, enabled bool,
) *quarkv1alpha1.NetAppStoragePool {
	sp.Spec.GlobalAccess = enabled
	return sp
}

func (client *NVCClient) NetAppStoragePoolGlobalAccessUpdateAbort(
	storagePoolUUID string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	sp, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(
		client.nvcNamespace).Get(
		context.TODO(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return nil, err
	}
	if sp.Status.GlobalAccess == sp.Spec.GlobalAccess {
		return nil, GlobalAccessUpdateAbortNoUpdateError{}
	}
	sp.Spec.GlobalAccess = sp.Status.GlobalAccess
	if nvc.HasFinalizer(sp.Finalizers,
		quarkconfig.ADClientAddressFinalizer) && sp.Status.ADClientIP == "" {
		// We're in the process of first time enablement; mark that we're
		// aborting
		sp.Spec.GlobalAccessAborting = true
	}
	return client.netAppStoragePoolUpdate(sp)
}

func (client *NVCClient) netAppStoragePoolStatusUpdate(storagePool *quarkv1alpha1.NetAppStoragePool) (*quarkv1alpha1.NetAppStoragePool, error) {
	storagePool, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).UpdateStatus(
		context.TODO(), storagePool, metav1.UpdateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"storagepool": storagePool.Name,
			"error":       err,
		}).Error("updating status failed.")

		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}

		return nil, err
	}
	return storagePool, nil
}

func (client *NVCClient) netappStoragePoolConditionUpdate(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	abortOperation bool,
	conditionTypes ...quarkv1alpha1.NetAppStoragePoolConditionType,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	updateStatus := false

	for _, conditionType := range conditionTypes {
		conditionPresent, retryErr := util.IsErrorConditionPresent(storagePool.Status.Conditions,
			conditionType)
		if conditionPresent {
			if abortOperation {
				storagePool.Status.Conditions = util.RemoveStoragePoolConditions(storagePool.Status.Conditions,
					conditionType)
				updateStatus = true
			} else if !retryErr {
				// Setting the ForceRetry flag in condition so that podrick can treat this condition as a retry error
				storagePool.Status.Conditions = util.SetForceRetryInErrorCondition(storagePool.Status.Conditions,
					conditionType)
				updateStatus = true
			}
		}
	}

	if updateStatus {
		return client.netAppStoragePoolStatusUpdate(storagePool)
	}
	return storagePool, nil
}

// Reset Machine Account password for SMB storagePool
func (client *NVCClient) NetAppStoragePoolSMBPasswordReset(storagePoolUUID string, user, password string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}
	secretName := util.GetVolPwdResetADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": storagePool.Name,
			"secret": secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBPasswordResetCount = storagePool.Spec.SMBPasswordResetCount + 1
	newStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	if err != nil {
		delErr := client.SecretDelete(secretName)
		if delErr != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
				"error":       delErr,
			}).Error("Failed to delete secret.")
		}

		return nil, err
	}
	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		newStoragePool, err = client.netappStoragePoolConditionUpdate(newStoragePool, false,
			quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
	}

	return newStoragePool, err
}

// Abort machine-account password reset operation.
// SDE should abort the password reset operation if it times out.
// Abort will delete the secret and avoid retrying operation.
func (client *NVCClient) NetAppStoragePoolADPasswordResetAbort(storagePoolUUID string) (sp *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolPwdResetADSecretName(storagePoolUUID)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	sp, err = client.NetAppStoragePoolGet(storagePoolUUID)
	if err != nil {
		return nil, err
	}

	sp.Spec.SMBPasswordResetCount = sp.Status.SMBPasswordResetCount
	sp, err = client.netAppStoragePoolUpdate(sp)

	// Cleanup up the error conditions
	if err == nil {
		sp, err = client.netappStoragePoolConditionUpdate(sp, true,
			quarkv1alpha1.NetAppStoragePoolSMBPasswordResetError)
	}

	return sp, err
}

// Update DNS IP for AD storagePool
func (client *NVCClient) NetAppStoragePoolADDNSUpdate(sp *quarkv1alpha1.NetAppStoragePool, dnsIPs []string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	// Returns error if the given IP address is not valid IPv4 address
	err = util.ValidateIPv4(dnsIPs)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = dnsIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolADDNSUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

// Update EncryptedCommunication for AD storagePool
func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdate(sp *quarkv1alpha1.NetAppStoragePool, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}
	return
}

// Update backup and security users from AD settings
func (client *NVCClient) NetAppStoragePoolSMBADBackupSecurityUsersUpdate(sp *quarkv1alpha1.NetAppStoragePool, backup_user_list []string, security_privilege_user_list []string) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	if !client.EnableCCPd || !client.EnableBackupSecurityUsers {
		return nil, fmt.Errorf("Cannot update users as CCPd flag or Backup/Security feature flag is disabled")
	}
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	err = client.validateNoDuplicates(backup_user_list, security_privilege_user_list)
	if err != nil {
		return nil, err
	}
	err = client.validateNames(backup_user_list, security_privilege_user_list)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBBackupUsers = backup_user_list
	storagePool.Spec.SMBSecurityPrivilegeUsers = security_privilege_user_list
	storagePool.Spec.SMBUsersUpdateCount = storagePool.Spec.SMBUsersUpdateCount + 1
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	return storagePool, nil
}

func (client *NVCClient) NetAppStoragePoolSMBADConfigurationUpdateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreate(sp *quarkv1alpha1.NetAppStoragePool, user, password string, ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration, SMBADConfiguration quarkv1alpha1.SMBADConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}

	// Returns error if the given IP address is not valid IPv4 address
	err = util.ValidateIPv4(ADDNSIPs)
	if err != nil {
		return nil, err
	}

	if client.NetAppStoragePoolSMBInitialized(storagePool) {
		if storagePool.Status.SMBServerConfiguration != SMBServerConfiguration || storagePool.Status.SMBADConfiguration != SMBADConfiguration ||
			!util.IsEqualStrings(storagePool.Status.ADDNSIPs, ADDNSIPs) {
			return nil, fmt.Errorf("updating CIFS Server config is not allowed: Server already congfigured")
		}
	}

	secretName := util.GetVolCreateADSecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}
	return
}

func (client *NVCClient) NetAppStoragePoolSMBConfigCreateAbort(sp *quarkv1alpha1.NetAppStoragePool) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	secretName := util.GetVolCreateADSecretName(sp.Name)
	err = client.SecretDelete(secretName)
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	storagePool.Spec.ADDNSIPs = storagePool.Status.ADDNSIPs
	storagePool.Spec.SMBServerConfiguration = storagePool.Status.SMBServerConfiguration
	storagePool.Spec.SMBADConfiguration = storagePool.Status.SMBADConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Cleanup up the error conditions
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, true,
			quarkv1alpha1.NetAppStoragePoolADDNSConfigurationError,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}

	return
}

func (client *NVCClient) NetAppStoragePoolSMBServerConfigurationUpdate(sp *quarkv1alpha1.NetAppStoragePool, user, password string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.NetAppStoragePoolGet(sp.Name)
	if err != nil {
		return nil, err
	}

	if storagePool.Status.SMBServerConfiguration.Netbios == "" {
		return nil, fmt.Errorf("updating CIFS Server config is not allowed: Server not yet congfigured")
	}

	// Domain and PreferredADSite are not allowed to change.
	if SMBServerConfiguration.Domain != storagePool.Status.SMBServerConfiguration.Domain ||
		SMBServerConfiguration.PreferredADSite != storagePool.Status.SMBServerConfiguration.PreferredADSite {
		return nil, fmt.Errorf("Domain and PreferredADSite are not allowed to change.")
	}

	if SMBServerConfiguration.Netbios == "" {
		return nil, fmt.Errorf("Netbios can't be empty.")
	}

	// If IsSMBServerConfigurationEqual is false then previous SMB config modify is failed or not completed.
	// In that case allow to continue modify the username and password
	if util.IsSMBServerConfigurationEqual(storagePool) &&
		SMBServerConfiguration.Netbios == storagePool.Spec.SMBServerConfiguration.Netbios &&
		SMBServerConfiguration.OrganizationUnit == storagePool.Spec.SMBServerConfiguration.OrganizationUnit {
		return nil, fmt.Errorf("invalid inputs. OrganizationUnit and Netbios are same as current configuration.")
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}

	secretName := util.GetSMBServerModifySecretName(storagePool.Name)
	_, err = client.createOrUpdateADSecretWithOwnerRef(nil, storagePool, secretName, user, password)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create secret.")
		return nil, err
	}

	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool, err = client.netAppStoragePoolUpdate(storagePool)

	// Update non-retry error condition to retry error condition so that podrick will act on this changes
	// Updating the condition after netAppStoragePoolUpdate so that the conditions will not get reverted back
	// by podrick before the netAppStoragePoolUpdate
	if err == nil {
		storagePool, err = client.netappStoragePoolConditionUpdate(storagePool, false,
			quarkv1alpha1.NetAppStoragePoolSMBServerConfigurationError)
	}

	return
}

// Validate availableZones, verify zone exists in availableZones, validate zone is non-empty
// for zonal storage pools in regional clusters.
func (client *NVCClient) validateZones(serviceLevel ServiceLevel, zone string, availableZones []string) (bool, error) {
	var err error

	switch serviceLevel {
	// For zonal SPs in regional cluster, only one availability zone should be passed.
	case ServiceLevelStandard:
		if zone == "" {
			return false, fmt.Errorf("specify target zone when deploying zonal storagePool in a regional cluster")
		}
		if len(availableZones) == quarkconfig.AvailableZonesGCP1P && zone == availableZones[0] {
			return true, nil
		} else if len(availableZones) != quarkconfig.AvailableZonesGCP1P {
			return false, fmt.Errorf("invalid number of availableZones, "+
				"expected %d, received %d",
				quarkconfig.AvailableZonesGCP1P, len(availableZones))
		} else if zone != availableZones[0] {
			return false, fmt.Errorf("target zone doesn't exist in availableZones, "+
				"expected %s, received %s",
				availableZones[0], zone)
		}
	case ServiceLevelRegionalStandard:
		if len(availableZones) == 0 {
			// We can use the default behavior; no reason to look up the cloud
			// provider
			return true, nil
		}
		// Need to know what cloud provider
		quarks, err := client.NetAppQuarkGetAll()
		if err != nil {
			return false, fmt.Errorf("unable to get Cloud provider, error retrieving quark CR: %v", err)
		}
		if len(quarks) == 0 {
			return false, fmt.Errorf("unable to get Cloud provider, quark CR not found")
		}
		quark := quarks[0]
		if quark.Spec.Cloud == quarkconfig.CloudGCP {
			if len(availableZones) != quarkconfig.AvailableZonesGCP {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesGCP, len(availableZones))
			}
			if !strings.Contains(strings.Join(availableZones, ","), zone) {
				return false, fmt.Errorf("target zone doesn't exist in availableZones, "+
					"expected %s, received %s",
					availableZones[0], zone)
			}
		} else if quark.Spec.Cloud == quarkconfig.CloudAzure {
			if len(availableZones) != quarkconfig.AvailableZonesAzure {
				return false, fmt.Errorf("invalid number of availableZones, "+
					"for Cloud provider, %s, expect %d, received %d",
					quark.Spec.Cloud, quarkconfig.AvailableZonesAzure, len(availableZones))
			}
		}
	}
	return true, err
}

// ConstructRegionalNetAppStoragePoolWithSMBConfig() does validations on availableZones before calling
// ConstructNetAppStoragePoolWithSMBConfig.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePoolWithSMBConfig(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	zonesOk, err := client.validateZones(serviceLevel, zone, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePoolWithSMBConfig(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel, ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones

	return storagePool, err
}

func (client *NVCClient) validateNoDuplicates(backup_user_list []string, security_privilege_user_list []string) (err error) {
	BackupUserVisitMap := make(map[string]bool, 0)
	for i := 0; i < len(backup_user_list); i++ {
		if BackupUserVisitMap[backup_user_list[i]] == true {
			return fmt.Errorf("Duplicate name %s present in Backup user list", backup_user_list[i])
		} else {
			BackupUserVisitMap[backup_user_list[i]] = true
		}
	}

	SecurityPrivilegeUserVisitMap := make(map[string]bool, 0)
	for i := 0; i < len(security_privilege_user_list); i++ {
		if SecurityPrivilegeUserVisitMap[security_privilege_user_list[i]] == true {
			return fmt.Errorf("Duplicate name %s present in Security privilege user list", security_privilege_user_list[i])
		} else {
			SecurityPrivilegeUserVisitMap[security_privilege_user_list[i]] = true
		}
	}

	return nil
}

func (client *NVCClient) validateNames(backup_user_list []string, security_privilege_user_list []string) (err error) {

	// Validation reference: Cloud Volumes Active directory name validations for backup and security privilege users.
	var max_len = 255
	invalid_chars := "\"/[]:;|=+*?<>@"
	for _, name := range backup_user_list {
		if strings.ContainsAny(name, invalid_chars) {
			return fmt.Errorf("Invalid characters in backup user %s, cannot contain any of %s", name, invalid_chars)
		}
		if len(name) > max_len {
			return fmt.Errorf("Length of name %s cannot be larger than %d", name, max_len)
		}
	}
	for _, name := range security_privilege_user_list {
		if strings.ContainsAny(name, invalid_chars) {
			return fmt.Errorf("Invalid characters in security privilege user %s, cannot contain any of %s", name, invalid_chars)
		}
		if len(name) > max_len {
			return fmt.Errorf("Length of name %s cannot be larger than %d", name, max_len)
		}
	}
	return nil
}

func (client *NVCClient) ValidateLdapConfig(ADDNSIPs []string, SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration) (err error) {
	ADDNSIPsString := strings.Join(ADDNSIPs, ",")
	if ADDNSIPsString == "" || SMBServerConfiguration.Domain == "" ||
		SMBServerConfiguration.Netbios == "" || SMBServerConfiguration.OrganizationUnit == "" {
		return fmt.Errorf("cannot create LDAP configuration: one or more empty SMB configuration parameters")
	}
	return nil
}

func (client *NVCClient) ConstructNetAppStoragePoolWithSMBConfig(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		log.Fatalf("Failed to construct StoragePool")
		return
	}

	err = util.ValidateIPv4(ADDNSIPs)
	if err != nil {
		return nil, err
	}

	err = nvc.ValidateADConfiguration(SMBServerConfiguration.Netbios, SMBServerConfiguration.Domain, SMBServerConfiguration.OrganizationUnit, SMBServerConfiguration.PreferredADSite)
	if err != nil {
		return nil, err
	}
	storagePool.Spec.SMBServerConfiguration = SMBServerConfiguration
	storagePool.Spec.ADDNSIPs = ADDNSIPs
	storagePool.Spec.SMBADConfiguration = SMBADConfiguration

	return storagePool, err
}

// calls ConstructRegionalNetAppStoragePoolWithSMBConfig() and sets LDAPEnabled to true in storagePool.Spec
func (client *NVCClient) ConstructNetAppStoragePoolWithSMBConfigAndLDAP(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
	ADDNSIPs []string,
	SMBServerConfiguration quarkv1alpha1.SMBServerConfiguration,
	SMBADConfiguration quarkv1alpha1.SMBADConfiguration,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	if !client.EnableCCPd || !client.EnableLdap {
		return nil, fmt.Errorf("cannot create LDAP configuration: CCPd flag or LDAP feature flag is disabled")
	}
	err = client.ValidateLdapConfig(ADDNSIPs, SMBServerConfiguration)
	if err != nil {
		return nil, err
	}

	// For regional storage pools (or) 1P config, we use 'ConstructRegionalNetAppStoragePool'
	if serviceLevel == ServiceLevelRegionalStandard || client.EnableGcp1p {
		storagePool, err = client.ConstructRegionalNetAppStoragePoolWithSMBConfig(ownerID,
			displayName, sizeString, storagePoolUUID, zone, availableZones, serviceLevel,
			ADDNSIPs, SMBServerConfiguration, SMBADConfiguration)
	} else {
		storagePool, err = client.ConstructNetAppStoragePoolWithSMBConfig(ownerID,
			displayName, sizeString, storagePoolUUID, serviceLevel, ADDNSIPs,
			SMBServerConfiguration, SMBADConfiguration)
	}
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}
	storagePool.Spec.LdapEnabled = true

	return storagePool, err
}

// ConstructRegionalNetAppStoragePool() does validations on availableZones before calling
// ConstructNetAppStoragePool.  It then sets Zone and AvailableZones fields of storagePool.Spec
func (client *NVCClient) ConstructRegionalNetAppStoragePool(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	zone string,
	availableZones []string,
	serviceLevel ServiceLevel,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	zonesOk, err := client.validateZones(serviceLevel, zone, availableZones)
	if !zonesOk {
		return nil, err
	}

	storagePool, err = client.ConstructNetAppStoragePool(ownerID, displayName, sizeString, storagePoolUUID, serviceLevel)
	if storagePool == nil || err != nil {
		return nil, fmt.Errorf("failed to construct StoragePool: %v", err)
	}

	// Set Zone and availableZones
	storagePool.Spec.Zone = zone
	storagePool.Spec.AvailableZones = availableZones

	return storagePool, err
}

// Constructs NetAppStoragePool definition.
// ownerID: OwnerID of the StoragePool.
// dispalyName: Name of the StoragePool.
// sizeString: StoragePool Size.
// storagePoolUUID: Unique ID of storagePool in k8s cluster.
// serviceLevel: Standard for alpha.
func (client *NVCClient) ConstructNetAppStoragePool(
	ownerID string,
	displayName string,
	sizeString string,
	storagePoolUUID string,
	serviceLevel ServiceLevel,
) (storagePool *quarkv1alpha1.NetAppStoragePool, err error) {
	defer func() {
		if e := recover(); e != nil {
			err = fmt.Errorf("invalid value for size")
		}
	}()

	// Validate service level
	switch serviceLevel {
	case ServiceLevelStandard, ServiceLevelRegionalStandard:
	default:
		return nil, fmt.Errorf("invalid value for service level")
	}

	// validate UUID
	_, err = uuid.FromString(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for UUID")
	}
	// Validate size
	size, err := resource.ParseQuantity(sizeString)
	if err != nil {
		return nil, fmt.Errorf("invalid value for size")
	}
	err = client.validateStoragePoolSize(size)
	if err != nil {
		return nil, err
	}

	// Validate ownerID
	if ownerID == "" {
		return nil, fmt.Errorf("storage pool OwnerID required")
	}

	storagePool = &quarkv1alpha1.NetAppStoragePool{
		metav1.TypeMeta{},
		metav1.ObjectMeta{
			Name: storagePoolUUID,
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel: storagePoolUUID,
				quarkconfig.OwnerIDLabel:     ownerID,
			},
		},
		quarkv1alpha1.NetAppStoragePoolSpec{
			Size:         size,
			DisplayName:  displayName,
			ServiceLevel: string(serviceLevel),
		},
		quarkv1alpha1.NetAppStoragePoolStatus{},
	}
	return storagePool, err
}

// Label a quark storage pool with SDE's storage pool uuid.
// This function expects 2 arguments .
// storagePoolUUID  is the name of the sp to label. sdeSpUUID is SP uuid assigned by SDE.
func (client *NVCClient) LabelNetAppStoragePool(storagePoolUUID, sdeSpUUID string) error {
	if sdeSpUUID == "" {
		return fmt.Errorf("provide non empty uuid of storage pool")
	}

	sp, err := client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.Background(), storagePoolUUID, metav1.GetOptions{})
	if err != nil {
		return err
	}
	if len(sp.Labels) == 0 {
		sp.Labels = make(map[string]string)
	}

	sp.Labels[quarkconfig.SDESpUUID] = sdeSpUUID

	_, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Update(context.Background(), sp, metav1.UpdateOptions{})

	return err
}

// optionsFn: Additional config for new features like cmek etc - clients can provide optional additional config by providing a
//
//	         variable number of options functions. An example option function is:
//	         func WithCmekConfig(cmekCfg *quarkv1alpha1.CMEKConfiguration) quarkv1alpha1.StoragePoolOptionsFn {
//					return func(additionalSPCfg *quarkv1alpha1.AdditionalStoragePoolConfig) {
//								additionalSPCfg.CMEKConfig = cmekCfg
//						   }
//	         }
func (client *NVCClient) NetAppStoragePoolCreate(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	user,
	password string,
	optionFns ...quarkv1alpha1.StoragePoolOptionsFn,
) (sp *quarkv1alpha1.NetAppStoragePool, err error) {

	var secretName string
	// operator allows creation of one quark CR only, only one element will be present in quarks list
	quarks, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, fmt.Errorf("failed to get Quark: %s", err.Error())
	}
	if len(quarks) == 0 {
		return nil, fmt.Errorf("no quarks found")
	}
	quark := quarks[0]
	// add quark version to create NetAppStoragePool only for "Production" & "TestVariant" configs
	switch strings.ToLower(quark.Spec.ControllerConfig.Variant) {
	case quarkconfig.ProductionVariant, quarkconfig.OnPremVariant:
		storagePool.Spec.QuarkVersion = quark.Status.Version
	}
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" {
		secretName = util.GetVolCreateADSecretName(storagePool.Name)
		_, err = client.createOrUpdateADSecret(secretName, user, password)
		if err != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"secret":      secretName,
			}).Error("Failed to create secret: ", err)
			return nil, err
		}
	}

	additionalConfig := quarkv1alpha1.AdditionalStoragePoolConfig{}
	for _, optionFn := range optionFns {
		optionFn(&additionalConfig)
	}

	var cmekConfigMap *corev1.ConfigMap
	// TODO-cmek add gcp1p flag whenever cmek feature is lightened
	if quark.Spec.EnableCMEK && (additionalConfig.CMEKConfig != nil) {
		// Check if a Storage Pool with the same UUID already exists. If it does do not create CMEK configmap as eventually the
		// Storage Pool would not be created as well.
		_, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Get(context.TODO(), storagePool.Name, metav1.GetOptions{})
		if err == nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
			}).Debug("storage pool with this name already exists", err)
			return nil, fmt.Errorf("storage pool with this name already exists")
		}

		// If the CMEKEmulatorConfig exists we need to set couple of additional fields in the CMEK ConfigMap for Emulator Testing purpose
		if cmekConfigMap, err = client.createCmekConfigMap(storagePool.Name, additionalConfig.CMEKConfig, additionalConfig.CMEKEmulatorConfig); err != nil {
			return nil, err
		}
		storagePool.Labels[quarkconfig.CMEKConfigMapLabel] = cmekConfigMap.Name
	}

	if quark.Spec.EnableGcp1p {
		if len(additionalConfig.IlbSubnets) == 0 {
			return nil, fmt.Errorf("failed to create Storage pool: no subnets are specified")
		}
		storagePool.Spec.IlbSubnets = additionalConfig.IlbSubnets
	}

	sp, err = client.quarkClient.QuarkV1alpha1().NetAppStoragePools(client.nvcNamespace).Create(context.TODO(), storagePool, metav1.CreateOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"secret":      secretName,
		}).Error("Failed to create storagePool:", err)
		return nil, err
	}

	// TODO-cmek add gcp1p flag whenever cmek feature is lightened
	if quark.Spec.EnableCMEK && (additionalConfig.CMEKConfig != nil) {
		cmekConfigMap.OwnerReferences = append(cmekConfigMap.OwnerReferences,
			*metav1.NewControllerRef(
				sp,
				schema.GroupVersionKind{
					Group:   quarkv1alpha1.SchemeGroupVersion.Group,
					Version: quarkv1alpha1.SchemeGroupVersion.Version,
					Kind:    "NetAppStoragePool",
				},
			),
		)
		_, err = client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Update(context.TODO(), cmekConfigMap, metav1.UpdateOptions{})
		if err != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"configMap":   cmekConfigMap.Name,
			}).Error("Failed to add owner reference to cmek config map: ", err)
			return nil, err
		}
	}

	return sp, nil
}

func (client *NVCClient) createCmekConfigMap(
	storagePoolName string,
	cmekConfig *quarkv1alpha1.CMEKConfiguration,
	cmekEmulatorConfig *quarkv1alpha1.CMEKEmulatorConfiguration,
) (*corev1.ConfigMap, error) {
	// Check all needed CMEK configs are present
	var errMsg string
	if cmekConfig == nil {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Cmek configs")
	} else if cmekConfig.Cloud != "gcp" {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Cloud=gcp")
	} else if len(strings.TrimSpace(cmekConfig.ProjectID)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Project Id")
	} else if len(strings.TrimSpace(cmekConfig.KeyRing)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Ring")
	} else if len(strings.TrimSpace(cmekConfig.KeyRingLocation)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Ring Location")
	} else if len(strings.TrimSpace(cmekConfig.KeyName)) == 0 {
		errMsg = fmt.Sprintf(CMEKConfigMapErr, "Key Name")
	}
	if len(strings.TrimSpace(errMsg)) != 0 {
		log.WithFields(log.Fields{
			"storagePool": storagePoolName,
		}).Error(errMsg)
		return nil, fmt.Errorf(errMsg)
	}
	cmekConfigMap := &corev1.ConfigMap{
		ObjectMeta: metav1.ObjectMeta{
			Name:      quarkconfig.CMEKConfigMapPrefix + storagePoolName,
			Namespace: client.nvcNamespace,
		},
		Data: map[string]string{
			quarkconfig.CMEKProtocolTag:            cmekConfig.Cloud,
			quarkconfig.CMEKProjectIDTag:           cmekConfig.ProjectID,
			quarkconfig.CMEKKeyRingTag:             cmekConfig.KeyRing,
			quarkconfig.CMEKKeyRingLocationTag:     cmekConfig.KeyRingLocation,
			quarkconfig.CMEKKeyNameTag:             cmekConfig.KeyName,
			quarkconfig.CMEKTimeoutTag:             quarkconfig.CMEKDefaultTimeout,
			quarkconfig.CMEKRootCACertMountPathTag: quarkconfig.RootCACertMountPath,
		},
	}

	// If CMEK Emulator config is set then set the Oauth and verify values as well
	if cmekEmulatorConfig != nil && strings.TrimSpace(cmekEmulatorConfig.OauthURL) != "" {
		cmekConfigMap.Data[quarkconfig.CMEKEmulatorOauthTag] = cmekEmulatorConfig.OauthURL
	}
	if cmekEmulatorConfig != nil && strings.TrimSpace(cmekEmulatorConfig.VerifyIpHost) != "" {
		cmekConfigMap.Data[quarkconfig.CMEKEmulatorVerifyIpHostTag] = cmekEmulatorConfig.VerifyIpHost
	}

	// If a valid log level is specified then set it in the CMEK configmap
	if len(strings.TrimSpace(cmekConfig.LogLevel)) != 0 {
		if isValidCMEKLogLevel(cmekConfig.LogLevel) {
			cmekConfigMap.Data[quarkconfig.CMEKLogLevelTag] = strings.ToLower(cmekConfig.LogLevel)
		} else {
			log.WithFields(log.Fields{
				"storagePool": storagePoolName,
			}).Debug("CMEK log level should be one of debug|error|critical|alert|info")
		}
	}

	cmekConfigMapCreated, err := client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Create(context.TODO(), cmekConfigMap, metav1.CreateOptions{})
	if errors.IsAlreadyExists(err) {
		// config map already exists most likely from a prior failed operation, replace it
		log.WithFields(log.Fields{
			"storagePool": storagePoolName,
			"configMap":   cmekConfigMap.Name,
		}).Debug("cmek config map exists already, replacing it: ", err)
		cmekConfigMapCreated, err = client.kubeClient.CoreV1().ConfigMaps(client.nvcNamespace).Update(context.TODO(), cmekConfigMap, metav1.UpdateOptions{})
	}
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePoolName,
			"configMap":   cmekConfigMap.Name,
		}).Error("Failed to create cmek config map: ", err)
		return nil, err
	}
	return cmekConfigMapCreated, nil
}

func (client *NVCClient) NetAppStoragePoolUpdateSubnets(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	ilbSubnets []string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	if len(ilbSubnets) == 0 {
		return nil, fmt.Errorf("no subnets specified, please specify one or more subnets")
	}
	storagePool.Spec.IlbSubnets = ilbSubnets
	updatedStoragePool, err := client.netAppStoragePoolUpdate(storagePool)
	return updatedStoragePool, err
}
